/******************************************************************************
 * Top contributors (to current version):
 *   Gereon Kremer
 *
 * This file is part of the cvc5 project.
 *
 * Copyright (c) 2009-2023 by the authors listed in the file AUTHORS
 * in the top-level source directory and their institutional affiliations.
 * All rights reserved.  See the file COPYING in the top-level source
 * directory for licensing information.
 * ****************************************************************************
 *
 * Option template for option modules.
 *
 * For each <module>_options.toml configuration file, mkoptions.py
 * expands this template and generates a <module>_options.cpp file.
 */
#include "options/parallel_options.h"

#include <iostream>

#include "base/check.h"
#include "options/option_exception.h"

namespace cvc5::internal::options {

// clang-format off
std::ostream& operator<<(std::ostream& os, CheckMode mode)
{
  switch(mode)
  {
    case CheckMode::STANDARD: return os << "standard";
    case CheckMode::FULL: return os << "full";
    default: Unreachable();
  }
  return os;
}
CheckMode stringToCheckMode(const std::string& optarg)
{
  if (optarg == "standard") return CheckMode::STANDARD;
  else if (optarg == "full") return CheckMode::FULL;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Partition check modes.
Available modes for --partition-check are:
+ standard (default)
  create partitions at standard checks
+ full
  create partitions at full checks
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --partition-check: `") +
                        optarg + "'.  Try --partition-check=help.");
}
std::ostream& operator<<(std::ostream& os, PartitionMode mode)
{
  switch(mode)
  {
    case PartitionMode::DECISION_SCATTER: return os << "decision-scatter";
    case PartitionMode::HEAP_SCATTER: return os << "heap-scatter";
    case PartitionMode::LEMMA_SCATTER: return os << "lemma-scatter";
    case PartitionMode::DECISION_CUBE: return os << "decision-cube";
    case PartitionMode::HEAP_CUBE: return os << "heap-cube";
    case PartitionMode::LEMMA_CUBE: return os << "lemma-cube";
    default: Unreachable();
  }
  return os;
}
PartitionMode stringToPartitionMode(const std::string& optarg)
{
  if (optarg == "decision-scatter") return PartitionMode::DECISION_SCATTER;
  else if (optarg == "heap-scatter") return PartitionMode::HEAP_SCATTER;
  else if (optarg == "lemma-scatter") return PartitionMode::LEMMA_SCATTER;
  else if (optarg == "decision-cube") return PartitionMode::DECISION_CUBE;
  else if (optarg == "heap-cube") return PartitionMode::HEAP_CUBE;
  else if (optarg == "lemma-cube") return PartitionMode::LEMMA_CUBE;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  Partition strategy modes.
Available modes for --partition-strategy are:
+ decision-scatter (default)
  For 4 partitions, creates partitions C1, !C1 & C2, !C1 & !C2 & C3, !C1 & !C2 &
  !C3, from decisions
+ heap-scatter
  For 4 partitions, creates partitions C1, !C1 & C2, !C1 & !C2 & C3, !C1 & !C2 &
  !C3, from heap
+ lemma-scatter
  For 4 partitions, creates partitions C1, !C1 & C2, !C1 & !C2 & C3, !C1 & !C2 &
  !C3, from lemmas
+ decision-cube
  Creates mutually exclusive cubes from the decisions in the SAT solver.
+ heap-cube
  Creates mutually exclusive cubes from the order heap in the SAT solver.
+ lemma-cube
  Creates mutually exclusive cubes from the lemmas sent to the SAT solver.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --partition-strategy: `") +
                        optarg + "'.  Try --partition-strategy=help.");
}
std::ostream& operator<<(std::ostream& os, PartitionWhenMode mode)
{
  switch(mode)
  {
    case PartitionWhenMode::TLIMIT: return os << "tlimit";
    case PartitionWhenMode::CLIMIT: return os << "climit";
    default: Unreachable();
  }
  return os;
}
PartitionWhenMode stringToPartitionWhenMode(const std::string& optarg)
{
  if (optarg == "tlimit") return PartitionWhenMode::TLIMIT;
  else if (optarg == "climit") return PartitionWhenMode::CLIMIT;
  else if (optarg == "help")
  {
    std::cerr << R"FOOBAR(
  PartitionWhen modes.
Available modes for --partition-when are:
+ tlimit (default)
  Partition when the time limit is exceeded.
+ climit
  Partition when number of checks is exceeded.
)FOOBAR";
    std::exit(1);
  }
  throw OptionException(std::string("unknown option for --partition-when: `") +
                        optarg + "'.  Try --partition-when=help.");
}
// clang-format on

}  // namespace cvc5::internal::options
